/******************************************************************************
* File name     : Dymamic Programming Module
* Date          : December 6, 2005
* Note          : The three functions (dpMatch_distance, dpMatch_threshold, 
*               : dpMatch_constraint) compute the DTW distance between given 
*               : two sequences. 
******************************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "dpmatch.h"

/******************************************************************************
* Function name : dpMatch_malloc
* Date          : December 6, 2005
* Input         : size_t      buffer_size1  ... Length of Sequence #1
* Input         : size_t      buffer_size2  ... Length of Sequence #2
* Output        : dp_match_t * != NULL      ... Address of the management table
*               :              == NULL      ... Memory allocation failure
* Note          : 
******************************************************************************/

dp_match_t* dpMatch_malloc(
  size_t      buffer_size1,
  size_t      buffer_size2
)
{

  dp_match_t* self;                    /* Management table                   */
  struct _dp_match_opt_value *buffer;  /* Results of distance computation    */
  size_t total_buffer_size;            /* Total buffer size                  */
  int *buf_integer;                    /* Warping scope                      */

  /*
   * Memory allocation for the management table
   */
  self = (dp_match_t *)malloc(sizeof(dp_match_t));
  if(self == NULL)
  {
    printf("Out of memory in dpMatch_malloc\n");
    exit(1);
  }

  /*
   * For keeping the results of distance computations
   */
  total_buffer_size = (buffer_size2 + 1)*2;
  buffer = (struct _dp_match_opt_value *)malloc(
    sizeof(struct _dp_match_opt_value) * total_buffer_size);
  if(buffer == NULL)
  {
    free(self);
    printf("Out of memory in dpMatch_malloc\n");
    exit(1);
  }
  self->result1 = buffer;      /* space needed is (buffer_size2 + 1)        */
  self->result2 = buffer + buffer_size2 + 1;    /* (buffer_size2 + 1)       */

  /*
   * Set the size of result1 and result2
   */
  self->buffer_size = buffer_size2;

  /*
   * For keeping information about the warping scope
   */
  buf_integer = (int *)malloc(sizeof(int) * buffer_size1 * 2);  
  if (buf_integer == NULL)
  {
    free(self->result1);
    free(self);
    printf("Out of memory in dpMatch_malloc\n");
    exit(1);
  }
  self->start_point = buf_integer;
  self->end_point = buf_integer + buffer_size1;

  return self;
}

/******************************************************************************
* Function name : dpMatch_initialization
* Date          : December 6, 2005
* Input         : dp_match_t* self          ... Address of the management table
* Input         : int         len1          ... Length of Sequence #1
* Input         : int         len2          ... Length of Sequence #2
* Note          : Set the starting and end points of the warping scope
******************************************************************************/

void dpMatch_initialization(
  dp_match_t* self,
  int         len1,
  int         len2
)
{

  float ratio, r1, r2;
  int w, i, *sp, *ep;
  int len1m, len2m, s1, s2, ii;

  /*
   * Set the sequence lengths
   */
  dpMatch_set_sequence_length(self, len1, len2);
  sp = self->start_point;
  ep = self->end_point;

  /*
   * Set the warping scope
   */
  ratio = (float)len2 / (float)len1;
#if ITAKURA
  /* Set warping path (Itakura band) */
  w = ITAKURA_WIDTH ;
  r1 = ratio / (float)w;
  r2 = ratio * w;
  len1m = len1 - 1;
  len2m = len2 - 1;
  s1 = (int)(len1 / (float)(w+1));
  s2 = len1 - s1;
  for (i=0 ; i<s2 ; i++)
    sp[i] = (int)floor(i * r1);
  for (i=len1m ; i>=s2 ; i--){
    ii = len1m - i;
    sp[i] = (int)floor((float)len2m - (float)(ii * r2));
  }
  for (i=0 ; i<s1 ; i++)
    ep[i] = (int)ceil(i * r2);
  for (i=len1m ; i>=s1 ; i--){
    ii = len1m - i;
    ep[i] = (int)ceil((float)len2m - (float)(ii * r1));
  }
  if (ep[0] < sp[1]-1)
    ep[0] = sp[1]-1;
  if (sp[len1m] > ep[len1m-1]+1)
    sp[len1m] = ep[len1m-1]+1;
#else
  /* Set warping path (Sacoe-Chiba band) */
  w = (int)((float)len2 * SC_WIDTH);
  for (i=0 ; i<len1 ; i++){
    sp[i] = (int)floor(i * ratio) - w;
    if (sp[i] < 0)
      sp[i] = 0;
    ep[i] = (int)ceil(i * ratio) + w;
    if (ep[i] >= len2)
      ep[i] = len2 - 1;
  }
#endif
  /*
  for (i=0 ; i<len1 ; i++)
  printf("%d %d %d\n", i, sp[i], ep[i]);
  */
}

/******************************************************************************
* Function name : dpMatch_clear
* Date          : December 6, 2005
* Input         : dp_match_t* self          ... Address of the management table
* Note          : Memory release
******************************************************************************/

void dpMatch_clear(
  dp_match_t* self
)
{

  free(self->start_point);
  free(self->result1);
  free(self);
}


/******************************************************************************
* Function name : dpMatch_distance
* Date          : December 6, 2005
* Input         : dp_match_t* self          ... Address of the management table
* Input         : float       *sequence1    ... Sequence #1
* Input         : float       *sequence2    ... Sequence #2
* Output        : float                     ... distance value
* Note          : Distance computation based on dynamic programming
*               :      seq1
*               :     +-----------
*               :    s|  |
*               :    e|  |  Compute distances from upper to lower
*               :    q|  V
*               :    2|
*               : The result of the distance computation of the grid point
*               : (i,j) is stored in `res_cur[j]'. `res_pri[j]' is the result 
*               : of (i-1,j). The result of the grid point (i,j) is: 
*               : 
*               : res_cur[j] = D(sequence1[i], sequence2[j]) 
*               :                + min(res_pri[j-1], res_cur[j-1], res_pri[j])
*               : 
******************************************************************************/

float dpMatch_distance(
  dp_match_t* self,
  float       *sequence1,
  float       *sequence2
)
{

  int    length1;                    /* Length of Sequence #1                */
  int    length2;                    /* Length of Sequence #2                */

  float       dist_cur;              /* Distance value of the current grid   */
  float       dist_opt;              /* Optimum distance of the current grid */
  long int    pos1;                    /* Cursor of Sequence #1              */
  long int    pos2;                    /* Cursor of Sequence #2              */

  struct _dp_match_opt_value *res_cur; /* Results of the current column      */
  struct _dp_match_opt_value *res_pri; /* Results of the prior (past) column */
  struct _dp_match_opt_value *swap;    /* For pointer-swapping               */

  float      *sp1;                     /* Pointer for Sequence #1            */
  float      *sp2;                     /* Pointer for Sequence #2            */

  /*
   * Initialization
   */
  res_cur = self->result1;
  res_pri = self->result2;
  length1 = self->seq1_length;
  length2 = self->seq2_length;

  /*
   * Compute the results of the first column
   */
  res_pri[0].optimal_value = CALC_DISTANCE(sequence1[0], sequence2[0])
                               * DIAGONAL_WEIGHT;

  for(pos2 = 1, sp2 = sequence2 + 1; pos2 < length2; pos2++, sp2++)
  {
    /*
     * Only use the vertical direction for the first column
     */
    res_pri[pos2].optimal_value = res_pri[pos2 - 1].optimal_value
      + CALC_DISTANCE(sequence1[0], sequence2[pos2]);
  }

  /*
   * From the second column to the end column
   */
  for(sp1 = sequence1 + 1, pos1 = 1; pos1 < length1; sp1++, pos1++)
  {

    for(sp2 = sequence2, pos2 = 0; pos2 < length2; sp2++, pos2++)
    {

      /*
       * Check the three directions, diagonal, vertical, horizontal ones, 
       * then choose the smallest among these values. 
       */
      res_cur[pos2].optimal_value = INITIALIZE_VALUE;
                                 /* Initialize the distance value            */
      dist_cur = CALC_DISTANCE(*sp1, *sp2);
                                 /* Compute the distance of the current grid */

      if(pos2 > 0)
      { 
	/*
	 * Diagonal direction
	 */
        res_cur[pos2].optimal_value 
            = res_pri[pos2 - 1].optimal_value + dist_cur * DIAGONAL_WEIGHT;

	/*
	 * Vertical direction
	 */
        dist_opt = res_cur[pos2 - 1].optimal_value + dist_cur;
        if(res_cur[pos2].optimal_value > dist_opt)
          res_cur[pos2].optimal_value = dist_opt;
      }

      /*
       * Horizontal direction
       */
      dist_opt = res_pri[pos2].optimal_value + dist_cur;
      if(res_cur[pos2].optimal_value > dist_opt)
        res_cur[pos2].optimal_value = dist_opt;

    }

  /*
   * Swap the results of the distance computation
   */
    swap    = res_cur;
    res_cur = res_pri;
    res_pri = swap;
  }

  /*
   * Report the final result
   */
  return res_pri[pos2 - 1].optimal_value;

}

/******************************************************************************
* Function name : dpMatch_threshold
* Date          : December 6, 2005
* Input         : dp_match_t* self          ... Address of the management table
* Input         : float       *sequence1    ... Sequence #1
* Input         : float       *sequence2    ... Sequence #2
* Input         : float       threshold     ... threshold for EarlyStopping
* Output        : float                     ... distance value
* Note          : Distance computation using EarlyStopping
******************************************************************************/

float dpMatch_threshold(
  dp_match_t* self,
  float       *sequence1,
  float       *sequence2,
  float       threshold
)
{

  int         length1;               /* Length of Sequence #1                */
  int         length2;               /* Length of Sequence #2                */
  int         *warp_sp;              /* Starting points of the warping scope */
  int         *warp_ep;              /* End points of the warping scope      */
  long int    start_point;           /* Starting point of the current column */
  long int    start_point_prior;     /* Starting point of the prior column   */
  long int    end_point;             /* End point of the current column      */
  long int    end_point_prior;       /* End point of the prior column        */

  float       dist_cur;              /* Distance value of the current grid   */
  float       dist_opt;              /* Optimum distance of the current grid */
  long int    pos1;                    /* Cursor of Sequence #1              */
  long int    pos2;                    /* Cursor of Sequence #2              */

  struct _dp_match_opt_value *res_cur; /* Results of the current column      */
  struct _dp_match_opt_value *res_pri; /* Results of the prior (past) column */
  struct _dp_match_opt_value *swap;    /* For pointer-swapping               */

  float      *sp1;                     /* Pointer for Sequence #1            */
  float      *sp2;                     /* Pointer for Sequence #2            */

  /*
   * Initialization
   */
  res_cur = self->result1;
  res_pri = self->result2;
  length1 = self->seq1_length;
  length2 = self->seq2_length;
  warp_sp = self->start_point;
  warp_ep = self->end_point;

  /*
   * Compute the results of the first column
   */
  res_pri[0].optimal_value = CALC_DISTANCE(sequence1[0], sequence2[0])
                               * DIAGONAL_WEIGHT;

  end_point = warp_ep[0]+1;
  for(pos2 = 1, sp2 = sequence2 + 1; pos2 < end_point ; pos2++, sp2++)
  {
    /*
     * Only use the vertical direction for the first column
     */
    res_pri[pos2].optimal_value = res_pri[pos2 - 1].optimal_value
      + CALC_DISTANCE(sequence1[0], sequence2[pos2]);

    if (threshold < res_pri[pos2].optimal_value)
      break;
  }
  end_point_prior = pos2;
  start_point_prior = 0;

  /*
   * From the second column to the end column
   */
  for(sp1 = sequence1 + 1, pos1 = 1; pos1 < length1; sp1++, pos1++)
  {

    start_point = warp_sp[pos1];
    if (start_point < start_point_prior)
      start_point = start_point_prior;
    end_point = warp_ep[pos1]+1;

    for(sp2 = sequence2 + start_point, pos2 = start_point;
        pos2 < end_point;
        sp2++, pos2++)
    {

      /*
       * Check the three directions, diagonal, vertical, horizontal ones, 
       * then choose the smallest among these values. 
       */
      res_cur[pos2].optimal_value = INITIALIZE_VALUE;
                                 /* Initialize the distance value            */
      dist_cur = CALC_DISTANCE(*sp1, *sp2);
                                 /* Compute the distance of the current grid */

      /*
       * Diagonal direction
       */
      if(pos2 > start_point_prior && end_point_prior >= pos2)
        res_cur[pos2].optimal_value 
            = res_pri[pos2 - 1].optimal_value + dist_cur * DIAGONAL_WEIGHT;

      /*
       * Vertical direction
       */
      if(pos2 > start_point)
      {
        dist_opt = res_cur[pos2 - 1].optimal_value + dist_cur;
        if(res_cur[pos2].optimal_value > dist_opt)
          res_cur[pos2].optimal_value = dist_opt;
      }

      /*
       * Horizontal direction
       */
      if(end_point_prior > pos2)
      {
        dist_opt = res_pri[pos2].optimal_value + dist_cur;
        if(res_cur[pos2].optimal_value > dist_opt)
          res_cur[pos2].optimal_value = dist_opt;
      }
      else
      {
	if (threshold < res_cur[pos2].optimal_value)
	  if (pos1 != length1 - 1)
	    {
	      pos2++;
	      break;
	    }
      }
    }

    /*
     * Check available grid points in the current column
     */
    end_point = pos2;
    for (pos2 = start_point ; pos2 < end_point ; pos2++)
    {
      if (threshold >= res_cur[pos2].optimal_value)
      {
	start_point_prior = pos2;
	break;
      }
    }
    if (pos2 == end_point)
      return res_cur[pos2-1].optimal_value;
    for (pos2 = end_point - 1 ; pos2 >= start_point ; pos2--)
    {
      if (threshold >= res_cur[pos2].optimal_value)
      {
	end_point_prior = pos2 + 1;
	break;
      }
    }

  /*
   * Swap the results of the distance computation
   */
    swap    = res_cur;
    res_cur = res_pri;
    res_pri = swap;

  }

  /*
   * Report the final result
   */
  return res_pri[length2 - 1].optimal_value;

}

/******************************************************************************
* Function name : dpMatch_constraint
* Date          : December 6, 2005
* Input         : dp_match_t* self          ... Address of the management table
* Input         : float       *sequence1    ... Sequence #1
* Input         : float       *sequence2    ... Sequence #2
* Output        : float                     ... distance value
* Note          : Distance computation with a warping scope
******************************************************************************/

float dpMatch_constraint(
  dp_match_t* self,
  float       *sequence1,
  float       *sequence2
)
{

  int         length1;               /* Length of Sequence #1                */
  int         length2;               /* Length of Sequence #2                */
  int         *warp_sp;              /* Starting points of the warping scope */
  int         *warp_ep;              /* End points of the warping scope      */
  long int    start_point;           /* Starting point of the current column */
  long int    start_point_prior;     /* Starting point of the prior column   */
  long int    end_point;             /* End point of the current column      */
  long int    end_point_prior;       /* End point of the prior column        */

  float       dist_cur;              /* Distance value of the current grid   */
  float       dist_opt;              /* Optimum distance of the current grid */
  long int    pos1;                    /* Cursor of Sequence #1              */
  long int    pos2;                    /* Cursor of Sequence #2              */

  struct _dp_match_opt_value *res_cur; /* Results of the current column      */
  struct _dp_match_opt_value *res_pri; /* Results of the prior (past) column */
  struct _dp_match_opt_value *swap;    /* For pointer-swapping               */

  float      *sp1;                     /* Pointer for Sequence #1            */
  float      *sp2;                     /* Pointer for Sequence #2            */

  /*
   * Initialization
   */
  res_cur = self->result1;
  res_pri = self->result2;
  length1 = self->seq1_length;
  length2 = self->seq2_length;
  warp_sp = self->start_point;
  warp_ep = self->end_point;

  /*
   * Compute the results of the first column
   */
  res_pri[0].optimal_value = CALC_DISTANCE(sequence1[0], sequence2[0])
                               * DIAGONAL_WEIGHT;

  end_point = warp_ep[0]+1;
  for(pos2 = 1, sp2 = sequence2 + 1; pos2 < end_point ; pos2++, sp2++)
  {
    /*
     * Only use the vertical direction for the first column
     */
    res_pri[pos2].optimal_value = res_pri[pos2 - 1].optimal_value
      + CALC_DISTANCE(sequence1[0], sequence2[pos2]);
  }
  end_point_prior = end_point;
  start_point_prior = 0;

  /*
   * From the second column to the end column
   */
  for(sp1 = sequence1 + 1, pos1 = 1; pos1 < length1; sp1++, pos1++)
  {

    start_point = warp_sp[pos1];
    end_point = warp_ep[pos1]+1;

    for(sp2 = sequence2 + start_point, pos2 = start_point;
        pos2 < end_point;
        sp2++, pos2++)
    {

      /*
       * Check the three directions, diagonal, vertical, horizontal ones, 
       * then choose the smallest among these values. 
       */
      res_cur[pos2].optimal_value = INITIALIZE_VALUE;
                                 /* Initialize the distance value            */
      dist_cur = CALC_DISTANCE(*sp1, *sp2);
                                 /* Compute the distance of the current grid */

      /*
       * Diagonal direction
       */
      if(pos2 > start_point_prior && end_point_prior >= pos2)
        res_cur[pos2].optimal_value 
            = res_pri[pos2 - 1].optimal_value + dist_cur * DIAGONAL_WEIGHT;

      /*
       * Vertical direction
       */
      if(pos2 > start_point)
      {
        dist_opt = res_cur[pos2 - 1].optimal_value + dist_cur;
        if(res_cur[pos2].optimal_value > dist_opt)
          res_cur[pos2].optimal_value = dist_opt;
      }

      /*
       * Horizontal direction
       */
      if(end_point_prior > pos2)
      {
        dist_opt = res_pri[pos2].optimal_value + dist_cur;
        if(res_cur[pos2].optimal_value > dist_opt)
          res_cur[pos2].optimal_value = dist_opt;
      }

    }

    start_point_prior = start_point;
    end_point_prior = end_point;

  /*
   * Swap the results of the distance computation
   */
    swap    = res_cur;
    res_cur = res_pri;
    res_pri = swap;

  }

  /*
   * Report the final result
   */
  return res_pri[length2 - 1].optimal_value;

}
